@@ -21,6 +21,7 @@
 ###############################################################################
 from Bio import SeqIO
 from flight.__init__ import __version__
+
 __author__ = "Rhys Newell"
 __copyright__ = "Copyright 2020"
 __credits__ = ["Rhys Newell"]
@@ -38,17 +39,10 @@ import os
 import datetime
 
 # Function imports
-import numpy as np
-import pandas as pd
-from numba import config, set_num_threads
-import threadpoolctl
 import warnings
 import imageio
 
 # Self imports
-from .binning import Binner
-from .cluster import Cluster
-import flight.utils as utils
 
 # Debug
 debug = {
@@ -59,6 +53,7 @@ debug = {
     5: logging.DEBUG
 }
 
+
 ###############################################################################
 ############################### - Exceptions - ################################
 
@@ -162,7 +157,6 @@ def main():
 
     fit_options.set_defaults(func=fit)
 
-
     bin_options = subparsers.add_parser(
         'bin',
         description='Perform UMAP and then HDBSCAN on array of variant depths',
@@ -245,7 +239,6 @@ def main():
                              dest="n_neighbors",
                              default=100)
 
-
     bin_options.add_argument(
         '--a_spread',
         help=
@@ -309,37 +302,34 @@ def main():
                              default=8)
     bin_options.set_defaults(func=bin)
 
-    vamb_options = subparsers.add_parser(
-        'vamb',
-        description='Bin out the results of vamb',
+    filter_options = subparsers.add_parser(
+        'filter',
+        description='Filter out contigs larger than the given base pair threshold',
         formatter_class=CustomHelpFormatter,
         epilog='''
-                                        ~ vamb ~
-            How to use vamb:
+                                        ~ filter ~
+            How to use filter:
 
-            flight vamb --reference assembly.fasta --clusters vamb_clusters.tsv
+            flight filter --reference assembly.fasta --min_size 200000
 
             ''')
 
-    vamb_options.add_argument('--reference',
+    filter_options.add_argument('--reference',
                               help='The assembly file to be binned',
                               dest='assembly')
 
-    vamb_options.add_argument('--clusters',
-                              help='The vamb clusters',
-                              dest='clusters')
 
-    vamb_options.add_argument('--min_size',
+    filter_options.add_argument('--min_size',
                               help='Minimum bin size',
                               dest='min_size',
                               default=200000)
 
-    vamb_options.add_argument('--output',
+    filter_options.add_argument('--output',
                               help='The output directory',
                               dest='output',
-                              default='vamb_bins/')
+                              default='filtered_contigs/')
 
-    vamb_options.set_defaults(func=vamb)
+    filter_options.set_defaults(func=filter)
     ###########################################################################
     # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
     # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Parsing input ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
@@ -370,19 +360,28 @@ def main():
 
         args.func(args)
 
+
 def fit(args):
     prefix = args.input.replace(".npy", "")
     os.environ["NUMEXPR_MAX_THREADS"] = args.threads
+    os.environ["NUMBA_NUM_THREADS" ]= args.threads
+    os.environ["MKL_NUM_THREADS"] = args.threads
+    os.environ["OPENBLAS_NUM_THREADS"] = args.threads
+    from .cluster import Cluster
+    import numpy as np
+    # os.environ["MKL_NUM_THREADS"] = args.threads
+    # os.environ["OPENBLAS_NUM_THREADS"] = args.threads
+
     if not args.precomputed:
         clusterer = Cluster(args.input,
-                           prefix,
-                           n_neighbors=int(args.n_neighbors),
-                           min_cluster_size=int(args.min_cluster_size),
-                           min_samples=int(args.min_samples),
-                           min_dist=float(args.min_dist),
-                           n_components=int(args.n_components),
-                           threads=int(args.threads),
-                           )
+                            prefix,
+                            n_neighbors=int(args.n_neighbors),
+                            min_cluster_size=int(args.min_cluster_size),
+                            min_samples=int(args.min_samples),
+                            min_dist=float(args.min_dist),
+                            n_components=int(args.n_components),
+                            threads=int(args.threads),
+                            )
         clusterer.fit_transform()
         clusterer.cluster()
         # clusterer.break_clusters()
@@ -392,14 +391,14 @@ def fit(args):
         np.save(prefix + '_separation.npy', clusterer.cluster_separation())
     else:
         clusterer = Cluster(args.input,
-                           prefix,
-                           n_neighbors=int(args.n_neighbors),
-                           min_cluster_size=int(args.min_cluster_size),
-                           min_samples=int(args.min_samples),
-                           scaler="none",
-                           precomputed=args.precomputed,
-                           threads=int(args.threads),
-                           )
+                            prefix,
+                            n_neighbors=int(args.n_neighbors),
+                            min_cluster_size=int(args.min_cluster_size),
+                            min_samples=int(args.min_samples),
+                            scaler="none",
+                            precomputed=args.precomputed,
+                            threads=int(args.threads),
+                            )
         clusterer.cluster_distances()
         clusterer.plot_distances()
         np.save(prefix + '_labels.npy', clusterer.labels())
@@ -407,8 +406,14 @@ def fit(args):
 
 def bin(args):
     prefix = args.output
-    os.environ["NUMEXPR_MAX_THREADS"] = "1"
-    set_num_threads(int(args.threads))
+    os.environ["NUMEXPR_MAX_THREADS"] = args.threads
+    os.environ["NUMBA_NUM_THREADS"] = args.threads
+    os.environ["MKL_NUM_THREADS"] = args.threads
+    os.environ["OPENBLAS_NUM_THREADS"] = args.threads
+    from .binning import Binner
+    import flight.utils as utils
+    import threadpoolctl
+
     if args.long_input is None and args.input is None:
         logging.warning("bin requires either short or longread coverage values.")
         sys.exit()
@@ -420,186 +425,175 @@ def bin(args):
         clusterer = Binner(args.input,
                            args.long_input,
                            args.kmer_frequencies,
-                           # args.variant_rates,
                            prefix,
                            args.assembly,
                            n_neighbors=int(args.n_neighbors),
-                           metric=args.metric,
-                           min_cluster_size=int(args.min_cluster_size),
                            min_contig_size=int(args.min_contig_size),
-                           min_samples=int(args.min_samples),
                            min_dist=float(args.min_dist),
-                           scaler=args.scaler,
-                           n_components=int(args.n_components),
-                           cluster_selection_method=args.cluster_selection_method,
                            threads=int(args.threads),
                            a=float(args.a),
                            b=float(args.b),
+                           initialization='spectral'
                            )
 
-        kwargs_write = {'fps':1.0, 'quantizer':'nq'}
+        kwargs_write = {'fps': 1.0, 'quantizer': 'nq'}
         plots = []
         with threadpoolctl.threadpool_limits(limits=int(args.threads), user_api='blas'):
-            
-            if clusterer.tnfs.values.shape[0] > int(args.n_neighbors):
-                # First pass quick TNF filter to speed up next steps
-                clusterer.update_parameters()
-                clusterer.filter()
-                if clusterer.tnfs[~clusterer.disconnected].values.shape[0] > int(args.n_neighbors)*5:
-                    # Second pass intersection filtering
-                    clusterer.update_parameters()
-                    found_disconnections = clusterer.fit_disconnect()
-                    # found_disconnections = True
-                    if clusterer.tnfs[~clusterer.disconnected][~clusterer.disconnected_intersected].values.shape[0] > int(args.n_neighbors) * 5 and found_disconnections:
-                        # Final fully filtered embedding to cluster on
-                        # clusterer.update_parameters()
-                        clusterer.fit_transform()
-                        clusterer.cluster()
+            with warnings.catch_warnings():
+                warnings.simplefilter("ignore")
 
-                        ## Plot limits
-                        x_min = min(clusterer.embeddings[:, 0]) - 5
-                        x_max = max(clusterer.embeddings[:, 0]) + 5
-                        y_min = min(clusterer.embeddings[:, 1]) - 5
-                        y_max = max(clusterer.embeddings[:, 1]) + 5
+                if clusterer.tnfs.values.shape[0] > int(args.n_neighbors):
+                    # First pass quick TNF filter to speed up next steps and remove large contigs that
+                    # are too distant from other contigs. These contigs tend to break UMAP results
+                    clusterer.filter()
+                    if clusterer.tnfs[~clusterer.disconnected].values.shape[0] > int(args.n_neighbors) * 5:
+                        # Second pass intersection filtering
+                        clusterer.update_parameters()
+                        found_disconnections = clusterer.fit_disconnect()
+                        if clusterer.tnfs[~clusterer.disconnected][~clusterer.disconnected_intersected].values.shape[
+                            0] > int(args.n_neighbors) * 2:
+                            # Final fully filtered embedding to cluster on
+                            clusterer.update_umap_params(clusterer.large_contigs[~clusterer.disconnected][~clusterer.disconnected_intersected].shape[0])
+                            if found_disconnections:
+                                clusterer.fit_transform(clusterer.large_contigs[~clusterer.disconnected][~clusterer.disconnected_intersected]['tid'],
+                                                        int(args.n_neighbors))
+                            clusterer.embeddings = clusterer.intersection_mapper.embedding_
 
-                        plots.append(utils.plot_for_offset(clusterer.embeddings, clusterer.clusterer.labels_, x_min, x_max, y_min, y_max, 0))
-                        clusterer.bin_contigs(args.assembly, int(args.min_bin_size))
+                            clusterer.min_cluster_size = 2
+                            logging.info("HDBSCAN - Performing initial clustering.")
+                            clusterer.labels = clusterer.iterative_clustering(clusterer.embeddings,
+                                                                              prediction_data=True,
+                                                                              allow_single_cluster=False,
+                                                                              double=True)
 
-                        n = 0
-                        old_tids = []
-                        logging.info("Performing iterative clustering with disconnections...")
-                        with warnings.catch_warnings():
-                            warnings.simplefilter("ignore")
-                            while n <= 5:
-                                
-                                clusterer.pairwise_distances()
-                                try:
-                                    max_bin_id = max(clusterer.bins.keys()) + 1
-                                except ValueError:
-                                    max_bin_id = 1
-                                    
-                                if n == 0 or old_tids != set(clusterer.unbinned_tids):
-                                    old_tids = set(clusterer.unbinned_tids)
-                                else:
-                                    
-                                    break  # nothing changed
-                                plots = clusterer.reembed_unbinned(clusterer.unbinned_tids, max_bin_id,
-                                                                plots, x_min, x_max, y_min, y_max, n+1, delete_unbinned=True)
+                            ## Plot limits
+                            x_min = min(clusterer.embeddings[:, 0]) - 10
+                            x_max = max(clusterer.embeddings[:, 0]) + 10
+                            y_min = min(clusterer.embeddings[:, 1]) - 10
+                            y_max = max(clusterer.embeddings[:, 1]) + 10
+
+                            plots.append(
+                                utils.plot_for_offset(clusterer.embeddings, clusterer.labels, x_min, x_max, y_min, y_max,
+                                                      0))
+                            clusterer.bin_contigs(args.assembly, int(args.min_bin_size))
+
+                            clusterer.plot(['contig_108_pilon', 'contig_1250_pilon',
+                                            'scaffold_1715_pilon', 'contig_1687_pilon', 'contig_1719_pilon', 'contig_1718_pilon'])
+
+
+                            logging.info("Reclustering individual bins.")
+
+                            clusterer.sort_bins()
+
+                            clusterer.reembed(clusterer.unbinned_tids,
+                                              max(clusterer.bins.keys()), plots,
+                                              x_min, x_max, y_min, y_max, 0, 100,
+                                              reembed=True,
+                                              delete_unbinned=True,
+                                              force=True)
+
+                            clusterer.sort_bins()
+                            # If after everything there are excessively large clusters hanging around
+                            # This is where we send them to turbo hell. This step is probably the main
+                            # reason Rosella won't work on eukaryotic genomes, if we made this step optional
+                            # then it might work on them but I don't have any good benchmarks to test on.
+                            n = 0
+                            while n <= 100:
+                                logging.debug("iteration: ", n)
+                                clusterer.overclustered = False  # large clusters
+                                plots, n = clusterer.pairwise_distances(plots, n,
+                                                                        x_min, x_max,
+                                                                        y_min, y_max,
+                                                                        size_only=True,
+                                                                        reembed=True)
+                                clusterer.sort_bins()
                                 n += 1
+                                if not clusterer.overclustered:
+                                    break  # no more clusters have broken
 
-
-                            plots = clusterer.reembed_unbinned(clusterer.unbinned_tids, max_bin_id,
-                                                                   plots, x_min, x_max, y_min, y_max, n+1,
-                                                                   delete_unbinned=True,
-                                                                   bin_unbinned=False) # First pass get bins
-                            clusterer.pairwise_distances(bin_unbinned=True) # Bin out large unbinned contigs
-                            try:
-                                max_bin_id = max(clusterer.bins.keys()) + 1
-                            except ValueError:
-                                max_bin_id = 1
-                            plots = clusterer.reembed_unbinned(clusterer.unbinned_tids, max_bin_id,
-                                                                   plots, x_min, x_max, y_min, y_max, n+1,
-                                                                   delete_unbinned=True,
-                                                                   bin_unbinned=True) # second pass get bins
-
-                        clusterer.bin_filtered(int(args.min_bin_size))
-                        clusterer.plot()
-                    elif not found_disconnections: # run clustering based off first round of embeddings
-                        clusterer.cluster()
-                        clusterer.bin_contigs(args.assembly, int(args.min_bin_size))
-                        ## Plot limits
-                        x_min = min(clusterer.embeddings[:, 0]) - 5
-                        x_max = max(clusterer.embeddings[:, 0]) + 5
-                        y_min = min(clusterer.embeddings[:, 1]) - 5
-                        y_max = max(clusterer.embeddings[:, 1]) + 5
-                        plots.append(utils.plot_for_offset(clusterer.embeddings, clusterer.clusterer.labels_, x_min, x_max, y_min, y_max, 0))
-
-
-                        n = 0
-                        old_tids = []
-                        logging.info("Performing iterative clustering...")
-                        with warnings.catch_warnings():
-                            warnings.simplefilter("ignore")
-                            while n <= 5:
-
-                                clusterer.pairwise_distances()
-                                try:
-                                    max_bin_id = max(clusterer.bins.keys()) + 1
-                                except ValueError:
-                                    max_bin_id = 1
-                                    
-                                if n == 0 or old_tids != set(clusterer.unbinned_tids):
-                                    old_tids = set(clusterer.unbinned_tids)
-                                else:
-
-                                    break  # nothing changed
-                                plots = clusterer.reembed_unbinned(clusterer.unbinned_tids,
-                                                                   max_bin_id,
-                                                                   plots, x_min, x_max, y_min, y_max, n+1, delete_unbinned=True)
+                            # This took so much refinement holy shit. This is what makes Rosella work
+                            # Each cluster is checked for internal metrics. If the metrics look bad then
+                            # Recluster -> re-embed -> recluster. If any of the new clusters look better
+                            # Than the previous clusters then take them instead (Within reason).
+                            # Kind of time consuming, could potentially be sped up with multiprocessing
+                            # but thread control might get a bit heckers.
+                            n = 0
+                            while n <= 100:
+                                clusterer.overclustered = False # large clusters
+                                plots, n = clusterer.pairwise_distances(plots, n,
+                                                                        x_min, x_max,
+                                                                        y_min, y_max,
+                                                                        reembed=True)
+                                clusterer.sort_bins()
                                 n += 1
+                                if not clusterer.overclustered:
+                                    break # no more clusters have broken
 
-                            plots = clusterer.reembed_unbinned(clusterer.unbinned_tids, max_bin_id,
-                                                                                           plots, x_min, x_max, y_min, y_max, n+1,
-                                                                                           delete_unbinned=True,
-                                                                                           bin_unbinned=False) # First pass get bins
-                            clusterer.pairwise_distances(bin_unbinned=True) # Bin out large unbinned contigs
-                            try:
-                                max_bin_id = max(clusterer.bins.keys()) + 1
-                            except ValueError:
-                                max_bin_id = 1
-                            plots = clusterer.reembed_unbinned(clusterer.unbinned_tids, max_bin_id,
-                                                                   plots, x_min, x_max, y_min, y_max, n+1,
-                                                                   delete_unbinned=True,
-                                                                   bin_unbinned=True) # second pass get bins
 
-                        clusterer.bin_filtered(int(args.min_bin_size))
-                        clusterer.plot()
+                            # THIS WILL BREAK BINS, Try not to use it
+                            # Quickly filter any busted contigs
+                            # These are just contigs that either belong by themselves or are
+                            # just noise that UMAP decided to put with other stuff. Only way to
+                            # get rid of them is to just use this smooth brain method
+                            # n = 0
+                            # while n <= 5:
+                            plots, n = clusterer.pairwise_distances(plots, n, x_min, x_max, y_min, y_max,
+                                                                    big_only=True)
+                            # #
+                            #     n += 1
+
+                            # Bin unfiltered but keep
+                            # clusterer.bin_filtered(1e6, keep_unbinned=True)
+                            # plots, n = clusterer.pairwise_distances(plots, n,
+                            #                                         x_min, x_max,
+                            #                                         y_min, y_max,
+                            #                                         dissolve=True)
+                            #
+                            # clusterer.reembed(clusterer.unbinned_tids,
+                            #                   max(clusterer.bins.keys()), plots,
+                            #                   x_min, x_max, y_min, y_max, n, 100,
+                            #                   reembed=True,
+                            #                   delete_unbinned=True,
+                            #                   force=True,
+                            #                   skip_clustering=True)
+
+
+                            clusterer.bin_filtered(int(args.min_bin_size), keep_unbinned=False, unbinned_only=False)
+                        else:
+                            clusterer.rescue_contigs(int(args.min_bin_size))
                     else:
                         clusterer.rescue_contigs(int(args.min_bin_size))
                 else:
                     clusterer.rescue_contigs(int(args.min_bin_size))
-            else:
-                clusterer.rescue_contigs(int(args.min_bin_size))
+            logging.debug("Writing bins...", len(clusterer.bins.keys()))
+            clusterer.write_bins(int(args.min_bin_size))
+            try:
+                imageio.mimsave(clusterer.path + '/UMAP_projections.gif', plots, fps=1)
+            except RuntimeError:  # no plotting has occurred due to no embedding
+                pass
 
-        clusterer.write_bins(int(args.min_bin_size))
-        imageio.mimsave(clusterer.path + '/UMAP_projections.gif', plots, fps=1)
 
-def vamb(args):
+def filter(args):
+    """
+    Filters out big contigs in an assembly and writes them to individual fasta files
+    """
     min_bin_size = int(args.min_size)
     prefix = args.output
     if not os.path.exists(prefix):
         os.makedirs(prefix)
 
-    bins = {}
-    with open(args.clusters, 'r') as vamb_file:
-        for line in vamb_file:
-            line = line.split()
-            try:
-                bins[line[0]].append(line[1])
-            except KeyError:
-                bins[line[0]] = [line[1]]
-
     assembly = SeqIO.to_dict(SeqIO.parse(args.assembly, "fasta"))
 
     logging.info("Writing bins...")
-    max_cluster_id = max(bins.keys())
-    for (bin, contigs) in bins.items():
-        if bin != -1:
-            # Calculate total bin size and check if it is larger than min_bin_size
-            bin_length = sum([len(assembly[contig].seq) for contig in contigs])
-            if bin_length >= min_bin_size:
-                with open(prefix + '/vamb_bin.' + str(bin) + '.fna', 'w') as f:
-                    for contig in contigs:
-                        write_contig(contig, assembly, f)
-
-        else:
-            # Get final bin value
-            max_cluster_id += 1
+    bin_id = 0
+    for (contig_name, record) in assembly.items():
+        if len(record.seq) >= min_bin_size:
+            bin_id += 1
             # Rescue any large unbinned contigs and put them in their own cluster
-            for contig in contigs:
-                if len(assembly[contig].seq) >= min_bin_size:
-                    with open(prefix + '/vamb_bin.' + str(max_cluster_id) + '.fna', 'w') as f:
-                        write_contig(contig, assembly, f)
+            with open(prefix + '/filtered_contig.' + str(bin_id) + '.fna', 'w') as f:
+                fasta = ">" + record.id + '\n'
+                fasta += str(record.seq) + '\n'
+                f.write(fasta)
 
 
 def write_contig(contig, assembly, f):
@@ -608,6 +602,7 @@ def write_contig(contig, assembly, f):
     fasta += str(seq.seq) + '\n'
     f.write(fasta)
 
+
 def phelp():
     print("""
 Usage:
@@ -643,9 +638,9 @@ class CustomHelpFormatter(argparse.HelpFormatter):
         h = action.help
         if '%(default)' not in action.help:
             if action.default != '' and \
-               action.default != [] and \
-               action.default != None \
-               and action.default != False:
+                    action.default != [] and \
+                    action.default != None \
+                    and action.default != False:
                 if action.default is not argparse.SUPPRESS:
                     defaulting_nargs = [
                         argparse.OPTIONAL, argparse.ZERO_OR_MORE
@@ -665,7 +660,5 @@ class CustomHelpFormatter(argparse.HelpFormatter):
         return ''.join([indent + line for line in text.splitlines(True)])
 
 
-
 if __name__ == '__main__':
-
-   sys.exit(main())
+    sys.exit(main())
